import setuptools
from os import path
# from setuptools.command.install import install
import accim


# class PostInstallCommand(install):
#     """Post-installation for installation mode."""
#     messages = []
# 
#     def run(self):
#         install.run(self)
#         # PUT YOUR POST-INSTALL SCRIPT HERE or CALL A FUNCTION
#         import os
#         import shutil
# 
#         # Creating a dictionary with IDD paths, IDD backups paths and the first line to modify
#         dict_lines = {
#             'EnergyPlus 25.1.0': [r"C:\EnergyPlusV25-1-0\Energy+.idd", r"C:\EnergyPlusV25-1-0\Energy+_backup.idd", 84912],
#             'EnergyPlus 24.2.0': [r"C:\EnergyPlusV24-2-0\Energy+.idd", r"C:\EnergyPlusV24-2-0\Energy+_backup.idd", 84877],
#             'EnergyPlus 24.1.0': [r"C:\EnergyPlusV24-1-0\Energy+.idd", r"C:\EnergyPlusV24-1-0\Energy+_backup.idd", 84380],
#             'EnergyPlus 23.2.0': [r"C:\EnergyPlusV23-2-0\Energy+.idd", r"C:\EnergyPlusV23-2-0\Energy+_backup.idd", 84284],
#             'EnergyPlus 23.1.0': [r"C:\EnergyPlusV23-1-0\Energy+.idd", r"C:\EnergyPlusV23-1-0\Energy+_backup.idd", 83101],
#             'EnergyPlus 22.2.0': [r"C:\EnergyPlusV22-2-0\Energy+.idd", r"C:\EnergyPlusV22-2-0\Energy+_backup.idd", 82643],
#             'EnergyPlus 22.1.0': [r"C:\EnergyPlusV22-1-0\Energy+.idd", r"C:\EnergyPlusV22-1-0\Energy+_backup.idd", 81925],
#             'EnergyPlus 9.6.0': [r"C:\EnergyPlusV9-6-0\Energy+.idd", r"C:\EnergyPlusV9-6-0\Energy+_backup.idd", 81798],
#             'EnergyPlus 9.5.0': [r"C:\EnergyPlusV9-5-0\Energy+.idd", r"C:\EnergyPlusV9-5-0\Energy+_backup.idd", 80493],
#             'EnergyPlus 9.4.0': [r"C:\EnergyPlusV9-4-0\Energy+.idd", r"C:\EnergyPlusV9-4-0\Energy+_backup.idd", 80526],
#             'EnergyPlus 9.3.0': [r"C:\EnergyPlusV9-3-0\Energy+.idd", r"C:\EnergyPlusV9-3-0\Energy+_backup.idd", 80397],
#             'EnergyPlus 9.2.0': [r"C:\EnergyPlusV9-2-0\Energy+.idd", r"C:\EnergyPlusV9-2-0\Energy+_backup.idd", 79641],
#             'EnergyPlus 9.1.0': [r"C:\EnergyPlusV9-1-0\Energy+.idd", r"C:\EnergyPlusV9-1-0\Energy+_backup.idd", 79116],
#             'EnergyPlus 9.0.0': [r"C:\EnergyPlusV9-0-0\Energy+.idd", r"C:\EnergyPlusV9-0-0\Energy+_backup.idd", 78635],
#         }
# 
#         new_lines_no = 7500
# 
#         # global messages
#         # messages = []
# 
#         # Iterating through EnergyPlus versions
#         for i in dict_lines:
#             # Renaming the original IDD file to keep it as a backup
#             try:
#                 os.rename(dict_lines[i][0], dict_lines[i][1])
#             except FileExistsError:
#                 message = f'The file Energy+_backup.idd already exists, therefore this script has already been run for {i}.'
#                 print(message)
#                 PostInstallCommand.messages.append(message)
#                 os.remove(dict_lines[i][0])
#                 # os.rename(dict_lines[i][1], dict_lines[i][0])
#                 shutil.copy(dict_lines[i][1], dict_lines[i][0])
#                 # continue
#             except FileNotFoundError:
#                 message = f'The file Energy+.idd has not been found, therefore {i} is not installed at default path.'
#                 print(message)
#                 PostInstallCommand.messages.append(message)
#                 continue
#             # Reading the content of the IDD
#             with open(dict_lines[i][1], 'r') as file:
#                 data = file.readlines()
#             # Making amendments in variable data
#             data[dict_lines[i][2]] = data[dict_lines[i][2]].replace(';', ',')
#             for j in range(503, new_lines_no + 1):
#                 if j == 503:
#                     line = dict_lines[i][2] + 1
#                 if j == new_lines_no:
#                     data.insert(line, f'A{j};\n')
#                 else:
#                     data.insert(line, f'A{j},\n')
#                 line = line + 1
# 
#             # Overwriting the IDD with the amendments in variable data
#             with open(dict_lines[i][0], 'w') as file:
#                 file.writelines(data)
#                 message = f"{i}'s IDD has been correctly modified. {new_lines_no} lines have been added to the object EnergyManagementSystem:Program."
#             print(message)
#             PostInstallCommand.messages.append(message)



this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='accim',
    version=accim.__version__,
    description='Transforms PMV-based into adaptive setpoint temperature EnergyPlus building energy models',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dsanchez-garcia/accim',
    author='Daniel Sánchez-García',
    author_email='daniel.sanchezgarcia@uca.es',
    license='GPL-3.0-or-later',
    license_files=('LICENSE',),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Scientific/Engineering'
        ],
    packages=setuptools.find_packages(),
    # package_data={
    #    '': [
    #        '*.csv',
    #        '*.idf',
    #        '*.epw',
    #        '*.eso',
    #        '*.jpg',
    #        '*.png',
    #        '*.xlsx',
    #        '*.xls',
    #        '*.ipynb'
    #         ]
    #    },
    include_package_data=True,
    python_requires='>=3.9',
    install_requires=[
        'scikit-learn',
        'eppy>=0.5.63',
        'pycountry',
        # 'geopy',
        'matplotlib',
        'numpy',
        'pandas',
        'datapackage',
        'unidecode',
        'besos',
        # 'shutil',
        # 'glob'
        ],
    # scripts=['bin/addAccis.py'],
    keywords=[
        'adaptive thermal comfort',
        'building energy model',
        'building performance simulation',
        'energy efficiency'
        ],
#     cmdclass={
#         'install': PostInstallCommand
#     }
    )

# print('IDD modifications:')
# print(PostInstallCommand.messages, sep='\n')


# accim - Adaptive-Comfort-Control-Implemented Model
# Copyright (C) 2021-2025 Daniel Sánchez-García

# accim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# accim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Run the function below to add the ACCIS.

This function transform fixed setpoint temperature
building energy models into adaptive setpoint temperature energy models
by adding the Adaptive Comfort Control Implementation Script (ACCIS)
"""

import pandas as pd
import besos.IDF_class
from accim import __version__

def addAccis(
    idf: besos.IDF_class = None,
    ScriptType: str = None,
    SupplyAirTempInputMethod: str = None,
    Output_type: str = None,
    Output_freqs: any = None,
    Output_keep_existing: bool = None,
    Output_take_dataframe: pd.DataFrame = None,
    EnergyPlus_version: str = None,
    TempCtrl: str = None,
    VRFschedule: str = 'On 24/7',
    verboseMode: bool = True,
    eer: float = 2,
    cop: float = 2.1,
    make_averages: bool = False,
    debugging: bool = False,
):

    """
    Adds the Adaptive-Comfort-Control Implementation Script, which is an EnergyManagementSystem
    script that applies adaptive setpoint temperatures to EnergyPlus building energy models.

    :param idf: The besos or eppy IDF class instance.
    :type idf: besos.IDF_class
    :param ScriptType: The default is None.
        'vrf_ac' for VRF system with full air-conditioning mode,
        'vrf_mm' for VRF system with mixed-mode,
        'ex_ac' for existing HVAC only with full air-conditioning mode,
        'ex_mm' for existing HVAC with mixed-mode.
    :type ScriptType: str
    :param SupplyAirTempInputMethod: The default is None.
        'supply air temperature' or 'temperature difference' to use such
        supply air temperature input method in the VRF system.
        Only used if vrf_ac or vrf_mm are used.
    :type SupplyAirTempInputMethod: str
    :param Output_type: The default is None.
        Can be 'standard', 'simplified', 'detailed' or 'custom'.
    :type Output_type: str
    :param Output_freqs: The default is None.
         A list containing the following strings:
         ['timestep', 'hourly', 'daily', 'monthly', 'runperiod']
    :type Output_freqs: list
    :param Output_keep_existing: The default is None.
        It is a boolean (True or False) to keep the existing Output:Variable objects or not.
    :type Output_keep_existing: bool
    :param Output_gen_dataframe: The default is None.
        It is a boolean (True or False) to generate a pandas DataFrame instance
        containing all Output:Variable objects.
    :type Output_gen_dataframe: bool
    :param Output_take_dataframe: It takes the pandas DataFrame previously generated
        with Output_gen_dataframe, which the user has filtered to keep only the rows
        related to the Output:Variable objects that need to be kept in the model.
    :type Output_take_dataframe: bool
    :param EnergyPlus_version: The default is None.
        Can be '9.1', '9.2', '9.3', '9.4', '9.5', '9.6', '22.1', '22.2', '23.1', '23.2', '24.1', '24.2', '25.1' or 'auto'.
    :type EnergyPlus_version: str
    :param TempCtrl: The default is None. Can be 'temp' or 'pmv'.
    :type TempCtrl: str
    :param verboseMode: True to print the process on screen. Default is True.
    :type verboseMode: bool
    :param eer: The energy efficiency ratio of the VRF system for each zone,
    added when using ScriptType vrf_mm or vrf_ac
    :type: eer: int
    :param cop: The coefficient of performance of the VRF system for each zone,
    added when using ScriptType vrf_mm or vrf_ac
    :type: cop: int
    :param make_averages: Used to make averages of hour-counting variables.
    :type make_averages: bool
    :param debugging: If True, an Output:EnergyManagementSystem object is used
        to generate the EDD file.
    :type debugging: bool
    :ivar arguments: A dictionary containing all arguments
    :ivar df_outputs: the pandas DataFrame instance created with argument ``Output_gen_dataframe``
    :ivar occupied_zones: A list containing all occupied zone names within the input idf.
    :ivar occupied_zones_original_name: A list containing all occupied zone original names within the input idf.
    :ivar windows_and_doors: A list containing all window and door names within the input idf.
    :ivar windows_and_doors_original_name:  A list containing all window and door original names within the input idf.
    """


    # import accim.sim.accim_Main_single_idf as accim_Main
    import accim.sim.accim_Main_single_idf as accim_Main

    import besos
    from besos.errors import InstallationError

    # IDF.setiddname(api_environment.EnergyPlusInputIddPath)
    # idf = IDF(api_environment.EnergyPlusInputIdfPath)

    fullScriptTypeList = [
        'vrf_ac',
        'vrf_mm',
        'ex_mm',
        'ex_ac',
    ]

    SupplyAirTempInputMethodList = [
        'supply air temperature',
        'temperature difference'
    ]

    fullOutputsTypeList = [
        'Standard',
        'standard',
        'Simplified',
        'simplified',
        'Detailed',
        'detailed',
        'Custom',
        'custom',
        # 'Show outputs',
        # 'show outputs'
    ]

    fullOutputsFreqList = [
        'Timestep',
        'timestep',
        'Hourly',
        'hourly',
        'Daily',
        'daily',
        'Monthly',
        'monthly',
        'Runperiod',
        'runperiod'
    ]

    fullEPversionsList = [
        '9.1',
        '9.2',
        '9.3',
        '9.4',
        '9.5',
        '9.6',
        '22.1',
        '22.2',
        '23.1',
        '23.2',
        '24.1',
        '24.2',
        '25.1',
    ]

    fullTempCtrllist = [
        'temperature',
        'temp',
        'pmv'
    ]

    if verboseMode:
        print(
            '\n--------------------------------------------------------'
            f'\nAdaptive-Comfort-Control-Implemented Model (ACCIM) v{__version__}'
            '\n--------------------------------------------------------'
            '\n\nThis tool allows to apply adaptive setpoint temperatures. '
            '\nFor further information, please read the documentation: '
            '\nhttps://accim.readthedocs.io/en/master/'
            '\nFor a visual understanding of the tool, please visit the following jupyter notebooks:'
            '\n-    Using addAccis() to apply adaptive setpoint temperatures'
            '\nhttps://accim.readthedocs.io/en/master/jupyter_notebooks/addAccis/using_addAccis.html'
            '\n-    Using rename_epw_files() to rename the EPWs for proper data analysis after simulation'
            '\nhttps://accim.readthedocs.io/en/master/jupyter_notebooks/rename_epw_files/using_rename_epw_files.html'
            '\n-    Using runEp() to directly run simulations with EnergyPlus'
            '\nhttps://accim.readthedocs.io/en/master/jupyter_notebooks/runEp/using_runEp.html'
            '\n-    Using the class Table() for data analysis'
            '\nhttps://accim.readthedocs.io/en/master/jupyter_notebooks/Table/using_Table.html'
            '\n-    Full example'
            '\nhttps://accim.readthedocs.io/en/master/jupyter_notebooks/full_example/full_example.html'
            '\n'
            '\nStarting with the process.'
        )



    if verboseMode:
        print('Basic input data:')
        # print(f'accim version: {accim.__version__}')
        print('ScriptType is: '+ScriptType)
    if ScriptType not in fullScriptTypeList:
        print('Valid ScriptTypes: ')
        print(fullScriptTypeList)
        raise ValueError(ScriptType + " is not a valid ScriptType. "
                                      "You must choose a ScriptType from the list above.")
    if 'vrf' in ScriptType.lower():
        if verboseMode:
            print('Supply Air Temperature Input Method is: '+SupplyAirTempInputMethod)
        if SupplyAirTempInputMethod not in SupplyAirTempInputMethodList:
            print('Valid Supply Air Temperature Input Methods: ')
            print(SupplyAirTempInputMethod)
            raise ValueError(SupplyAirTempInputMethod + " is not a valid Supply Air Temperature Input Method. "
                                          "You must choose a Supply Air Temperature Input Method from the list above.")
    if verboseMode:
        print('Output type is: ' + Output_type)
    if Output_type not in fullOutputsTypeList:
        print('Valid Output type: ')
        print(fullOutputsTypeList)
        raise ValueError(Output_type + " is not a valid Output. "
                                   "You must choose a Output from the list above.")
    if verboseMode:
        print('Output frequencies are: ')
        print(Output_freqs)
    if not (all(elem in fullOutputsFreqList for elem in Output_freqs)):
        print('Valid Output freqs: ')
        print(fullOutputsFreqList)
        raise ValueError('Some of the Output frequencies in '+Output_freqs + " is not a valid Output. "
                                   "All Output frequencies must be included in the list above.")
    if EnergyPlus_version is None:
        EnergyPlus_version = f'{idf.idd_version[0]}.{idf.idd_version[1]}'
    if verboseMode:
        print('EnergyPlus version is: '+EnergyPlus_version)
    if EnergyPlus_version not in fullEPversionsList:
        print('Valid EnergyPlus_version: ')
        print(fullEPversionsList)
        raise ValueError(EnergyPlus_version + " is not a valid EnergyPlus_version. "
                                              "You must choose a EnergyPlus_version"
                                              "from the list above.")
    if verboseMode:
        print('Temperature Control method is: '+TempCtrl)
    if TempCtrl not in fullTempCtrllist:
        print('Valid Temperature Control methods: ')
        print(fullTempCtrllist)
        raise ValueError(TempCtrl + " is not a valid Temperature Control method. "
                                              "You must choose a Temperature Control method"
                                              "from the list above.")
    arguments = {
        'ScriptType': ScriptType,
        'SupplyAirTempInputMethod': SupplyAirTempInputMethod,
        'Output_type': Output_type,
        'Output_freqs': Output_freqs,
        'Output_keep_existing': Output_keep_existing,
        'Output_take_dataframe': Output_take_dataframe,
        'EnergyPlus_version': EnergyPlus_version,
        'TempCtrl': TempCtrl,
        'verboseMode': verboseMode
    }


    if verboseMode:
        print('''\n=======================START OF GENERIC IDF FILE GENERATION PROCESS=======================\n''')
        print('Starting with file:')
        # print(idf)


    z = accim_Main.accimJob(
        idf_class_instance=idf,
        ScriptType=ScriptType,
        EnergyPlus_version=EnergyPlus_version,
        TempCtrl=TempCtrl,
        verboseMode=verboseMode
    )

    # self.occupied_zones = z.occupiedZones
    # self.occupied_zones_original_name = z.occupiedZones_orig
    # self.windows_and_doors = z.windownamelist
    # self.windows_and_doors_original_name = z.windownamelist_orig

    z.setComfFieldsPeople(EnergyPlus_version=EnergyPlus_version, TempCtrl=TempCtrl, verboseMode=verboseMode)

    if 'vrf' in ScriptType.lower():
        if TempCtrl.lower() == 'temperature' or TempCtrl.lower() == 'temp':
            z.addOpTempTherm(verboseMode=verboseMode)
        elif TempCtrl.lower() == 'pmv':
            z.setPMVsetpoint(verboseMode=verboseMode)
        z.addBaseSchedules(verboseMode=verboseMode)
        z.setAvailSchOn(verboseMode=verboseMode)
        z.addVRFsystemSch(verboseMode=verboseMode)
        z.addCurveObj(verboseMode=verboseMode)
        z.addDetHVACobj(
            EnergyPlus_version=EnergyPlus_version,
            verboseMode=verboseMode,
            SupplyAirTempInputMethod=SupplyAirTempInputMethod,
            eer=eer,
            cop=cop,
            VRFschedule=VRFschedule
        )
        if ScriptType.lower() == 'vrf_mm':
            z.checkVentIsOn(verboseMode=verboseMode)
        z.addForscriptSchVRFsystem(verboseMode=verboseMode)
    elif 'ex' in ScriptType.lower():
        # todo check if PMV can work with ex_ac
        z.addForscriptSchExistHVAC(verboseMode=verboseMode)

    z.addEMSProgramsBase(ScriptType=ScriptType, verboseMode=verboseMode)
    z.addEMSOutputVariableBase(ScriptType=ScriptType, verboseMode=verboseMode)
    z.addGlobVarList(ScriptType=ScriptType, verboseMode=verboseMode)
    z.addIntVarList(verboseMode=verboseMode)
    z.addEMSSensorsBase(ScriptType=ScriptType, verboseMode=verboseMode)
    z.addEMSActuatorsBase(ScriptType=ScriptType, verboseMode=verboseMode)

    if 'vrf' in ScriptType.lower():
        z.addEMSSensorsVRFsystem(ScriptType=ScriptType, verboseMode=verboseMode)
    elif ScriptType.lower() == 'ex_mm':
        z.addEMSSensorsExisHVAC(verboseMode=verboseMode)

    z.addEMSPCMBase(verboseMode=verboseMode)

    if make_averages:
        z.makeAverages(verboseMode=verboseMode)

    if Output_keep_existing == 'true':
        Output_keep_existing = True
    elif Output_keep_existing == 'false':
        Output_keep_existing = False
    if Output_keep_existing is True:
        pass
    else:
        z.removeExistingOutputVariables()

    if Output_type.lower() == 'simplified':
        z.addOutputVariablesSimplified(
            Outputs_freq=Output_freqs,
            TempCtrl=TempCtrl,
            verboseMode=verboseMode
        )
    elif Output_type.lower() == 'standard':
        z.addOutputVariablesStandard(
            Outputs_freq=Output_freqs,
            ScriptType=ScriptType,
            TempCtrl=TempCtrl,
            verboseMode=verboseMode
        )
    elif Output_type.lower() == 'detailed' or Output_type.lower() == 'custom':
        z.addOutputVariablesStandard(
            Outputs_freq=Output_freqs,
            ScriptType=ScriptType,
            TempCtrl=TempCtrl,
            verboseMode=verboseMode
        )
        z.addOutputVariablesDetailed(
            Outputs_freq=Output_freqs,
            verboseMode=verboseMode
        )
        if Output_type.lower() == 'custom':
            Output_gen_dataframe = False
            z.outputsSpecified()

    if Output_take_dataframe is not None:
        z.takeOutputDataFrame(
            idf_filename=idf.idfname.split('.idf')[0],
            df_outputs_in=Output_take_dataframe,
            verboseMode=verboseMode,
            singleidf=True,
        )

    z.removeDuplicatedOutputVariables()

    z.addControlFilesObjects(verboseMode=verboseMode)

    z.addOutputVariableDictionaryObject(verboseMode=verboseMode)

    if debugging:
        z.addOutputEnergyManagementSystem(verboseMode=verboseMode)

    if verboseMode:
        print('''\n=======================END OF OUTPUT IDF FILE GENERATION PROCESS=======================\n''')


    # self.idf = idf

def gen_outputs_df(
        idf: besos.IDF_class = None,
        ScriptType: str = None,
        Output_type: str = None,
        Output_freqs: any = None,
        Output_keep_existing: bool = None,
        TempCtrl: str = None,
        make_averages: bool = False,
        verboseMode: bool = False
) -> pd.DataFrame:
    """
    Returns a pandas.Dataframe instance which contains the Output:Variable objects in the idf.

    :param idf: The besos or eppy IDF class instance.
    :type idf: besos.IDF_class
    :param ScriptType: The default is None.
        'vrf_ac' for VRF system with full air-conditioning mode,
        'vrf_mm' for VRF system with mixed-mode,
        'ex_ac' for existing HVAC only with full air-conditioning mode,
        'ex_mm' for existing HVAC with mixed-mode.
    :type ScriptType: str
    :param Output_type: The default is None.
        Can be 'standard', 'simplified', 'detailed' or 'custom'.
    :type Output_type: str
    :param Output_freqs: The default is None.
         A list containing the following strings:
         ['timestep', 'hourly', 'daily', 'monthly', 'runperiod']
    :type Output_freqs: list
    :param Output_keep_existing: The default is None.
        It is a boolean (True or False) to keep the existing Output:Variable objects or not.
    :type Output_keep_existing: bool
    :param TempCtrl: The default is None. Can be 'temp' or 'pmv'.
    :type TempCtrl: str
    :param make_averages: Used to make averages of hour-counting variables.
    :type make_averages: bool
    :param verboseMode: True to print the process on screen. Default is True.
    :type verboseMode: bool
    :return: pandas.Dataframe instance which contains the Output:Variable objects in the idf
    """
    addAccis(
        idf=idf,
        ScriptType=ScriptType,
        Output_type=Output_type,
        Output_freqs=Output_freqs,
        Output_keep_existing=Output_keep_existing,
        TempCtrl=TempCtrl,
        make_averages=make_averages,
        SupplyAirTempInputMethod='temperature difference',
        verboseMode=verboseMode
    )
    
    import pandas as pd
    alloutputs = [
        output
        for output
        in idf.idfobjects['Output:Variable']
    ]
    df_outputs_temp = pd.DataFrame(columns=['key_value', 'variable_name', 'reporting_frequency', 'schedule_name'])
    for i in range(len(alloutputs)):
        df_outputs_temp.loc[i, 'key_value'] = alloutputs[i].Key_Value
        df_outputs_temp.loc[i, 'variable_name'] = alloutputs[i].Variable_Name
        df_outputs_temp.loc[i, 'reporting_frequency'] = alloutputs[i].Reporting_Frequency
        df_outputs_temp.loc[i, 'schedule_name'] = alloutputs[i].Schedule_Name

    return df_outputs_temp
    pass




def modifyAccis(
        idf,
        ComfStand: int = None,
        CAT: int = None,
        CATcoolOffset: float = 0,
        CATheatOffset: float = 0,
        ComfMod: float = None,
        SetpointAcc: float = 10000,
        CustAST_ACSTaul: float = 0,
        CustAST_ACSTall: float = 0,
        CustAST_AHSTaul: float = 0,
        CustAST_AHSTall: float = 0,
        CustAST_m: float = 0,
        CustAST_n: float = 0,
        CustAST_ACSToffset: float = 0,
        CustAST_AHSToffset: float = 0,
        CoolSeasonStart: any = 121,
        CoolSeasonEnd: any = 274,
        HVACmode: int = None,
        VentCtrl: int = None,
        MaxTempDiffVOF: float = 6,
        MinTempDiffVOF: float = 1,
        MultiplierVOF: float = 0.25,
        VSToffset: int = 0,
        MinOToffset: int = 50,
        MaxWindSpeed: int = 50,
        ASTtol: int = 0.1,
):
    """
    :param ComfStand: The default is None.
        '0 = ESP CTE;
        '1 = INT EN16798;
        '2 = INT ASHRAE55;
        '3 = JPN Rijal;
        '4 = CHN GBT50785 Cold;
        '5 = CHN GBT50785 HotMild;
        '6 = CHN Yang;
        '7 = IND IMAC C NV;
        '8 = IND IMAC C MM;
        '9 = IND IMAC R 7DRM;
        '10 = IND IMAC R 30DRM;
        '11 = IND Dhaka;
        '12 = ROM Udrea;
        '13 = AUS Williamson;
        '14 = AUS DeDear;
        '15 = BRA Rupp NV;
        '16 = BRA Rupp AC;
        '17 = MEX Oropeza Arid;
        '18 = MEX Oropeza DryTropic;
        '19 = MEX Oropeza Temperate;
        '20 = MEX Oropeza HumTropic;
        '21 = CHL Perez-Fargallo;
        '22 = INT ISO7730
    :type ComfStand: int
    :param CustAST_m: The m coefficient (slope) of custom model linear regression (mx+n)
    :type CustAST_m: float
    :param CustAST_n: The n coefficient of custom model linear regression (mx+n)
    :type CustAST_n: float
    :param CustAST_AHSToffset: The offset for heating setpoint from neutral temperature
        for the custom model linear regression. This value will be summed, therefore, it must be negative.
    :type CustAST_AHSToffset: float
    :param CustAST_ACSToffset: The offset for cooling setpoint from neutral temperature
        for the custom model linear regression. This value will be summed, therefore, it must be positive.
    :type CustAST_ACSToffset: float
    :param CustAST_ACSTaul: The value for the cooling setpoint applicability upper limit (ACSTaul).
    :type CustAST_ACSTaul: float
    :param CustAST_ACSTall: The value for the cooling setpoint applicability lower limit (ACSTall).
    :type CustAST_ACSTall: float
    :param CustAST_AHSTaul: The value for the heating setpoint applicability upper limit (AHSTaul).
    :type CustAST_AHSTaul: float
    :param CustAST_AHSTall: The value for the heating setpoint applicability lower limit (AHSTall).
    :type CustAST_AHSTall: float
    :param CAT: The default is None.
        (1 = CAT I; 2 = CAT II; 3 = CAT III; 80 = 80% ACCEPT; 85 = 85% ACCEPT; 90 = 90% ACCEPT)
    :type CAT: int
    :param CATcoolOffset: An offset to modify comfort models.
        This value is summed to the predefined cooling setpoint offset for the CAT value.
    :type CATcoolOffset: float
    :param CATheatOffset: An offset to modify comfort models.
        This value is summed to the predefined heating setpoint offset for the CAT value.
    :type CATheatOffset: float
    :param ComfMod: The default is None.
        (0/0.X = Static;
        1/1.X = Adaptive when applicable, otherwise relevant local static model;
        2 = Adaptive when applicable, otherwise relevant international static model
        3 = Adaptive when applicable, otherwise horizontal extention of adaptive setpoints)
    :type ComfMod: float
    :param SetpointAcc: A float. It is the number for the accuracy of the setpoint temperatures.
        For instance, if 2 was used, setpoints would be rounded to every half Celsius degree;
        if 10 was used, the setpoints would be rounded to the first decimal.
    :type SetpointAcc: float
    :param CoolSeasonStart: A date in format dd/mm, or the number of the day in the year.
        Defines when start the cooling season, only used in some static setpoint temperatures.
    :type CoolSeasonStart: any
    :param CoolSeasonEnd: A date in format dd/mm, or the number of the day in the year.
        Defines when ends the cooling season, only used in some static setpoint temperatures.
    :type CoolSeasonEnd: any
    :param HVACmode: The default is None.
        (0 = Fully Air-conditioned; 1 = Naturally ventilated; 2 = Mixed Mode)
    :type HVACmode: int
    :param VentCtrl: The default is None.
        (if HVACmode = 1:
        0 = Ventilates above neutral temperature;
        1 = Ventilates above upper comfort limit;
        if HVACmode = 2:
        0 = Ventilates above neutral temperature and fully opens doors and windows;
        1 = Ventilates above lower comfort limit and fully opens doors and windows;
        2 = Ventilates above neutral temperature and opens doors and windows based on the customised venting opening factor;
        3 = Ventilates above lower comfort limit and opens doors and windows based on the customised venting opening factor;
        )
    :type VentCtrl: int
    :param MaxTempDiffVOF: The maximum temperature difference for the Venting Opening Factor.
        Must be a number greater than 0.
    :type MaxTempDiffVOF: float
    :param MinTempDiffVOF: The minimum temperature difference for the Venting Opening Factor.
        Must be a number greater than 0 and smaller than the maximum temperature difference.
    :type MinTempDiffVOF: float
    :param MultiplierVOF: The multiplier for the modulation of the Venting Opening Factor.
        Must be a number between 0 and 1.
    :type MultiplierVOF: float
    :param VSToffset: The default is 0. Please refer to documentation.
    :type VSToffset: float
    :param MinOToffset: The default is 50. Please refer to documentation.
    :type MinOToffset: float
    :param MaxWindSpeed: The default is 50. Please refer to documentation.
    :type MaxWindSpeed: float
    :param ASTtol: The default is 0.1. Please refer to documentation.
    :type ASTtol: float
    """

    SetInputData = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                     program.Name == 'SetInputData'][0])
    SetVOFinputData = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                        program.Name == 'SetVOFinputData'][0])
    SetAST = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
               program.Name == 'SetAST'][0])
    ApplyCAT = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                 program.Name == 'ApplyCAT'][0])

    SetComfTemp = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                    program.Name == 'SetComfTemp'][0])
    SetAppLimits = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                     program.Name == 'SetAppLimits'][0])

    arguments_accis = {
        'ComfStand': ComfStand,
        'CAT': CAT,
        'CATcoolOffset': CATcoolOffset,
        'CATheatOffset': CATheatOffset,
        'ComfMod': ComfMod,
        'SetpointAcc': SetpointAcc,
        'CustAST_ACSTaul': CustAST_ACSTaul,
        'CustAST_ACSTall': CustAST_ACSTall,
        'CustAST_AHSTaul': CustAST_AHSTaul,
        'CustAST_AHSTall': CustAST_AHSTall,
        'CustAST_m': CustAST_m,
        'CustAST_n': CustAST_n,
        'CustAST_ACSToffset': CustAST_ACSToffset,
        'CustAST_AHSToffset': CustAST_AHSToffset,
        'CoolSeasonStart': CoolSeasonStart,
        'CoolSeasonEnd': CoolSeasonEnd,
        'HVACmode': HVACmode,
        'VentCtrl': VentCtrl,
        'MaxTempDiffVOF': MaxTempDiffVOF,
        'MinTempDiffVOF': MinTempDiffVOF,
        'MultiplierVOF': MultiplierVOF,
        'VSToffset': VSToffset,
        'MinOToffset': MinOToffset,
        'MaxWindSpeed': MaxWindSpeed,
        'ASTtol': ASTtol
    }

    while SetpointAcc < 0:
        raise ValueError('The value for SetpointAcc cannot be less than 0.')

    if type(CoolSeasonStart) is int:
        if CoolSeasonStart <= 365 and CoolSeasonStart > 0:
            pass
    elif type(CoolSeasonStart) is str:
        if len(CoolSeasonStart.split('/')) == 2:
            day = int(CoolSeasonStart.split('/')[0])
            month = int(CoolSeasonStart.split('/')[1])
            from datetime import date
            day_of_year = date(year=2007, month=month, day=day).timetuple().tm_yday
            CoolSeasonStart = day_of_year

    if type(CoolSeasonEnd) is int:
        if CoolSeasonEnd <= 365 and CoolSeasonEnd > 0:
            pass
    elif type(CoolSeasonEnd) is str:
        if len(CoolSeasonEnd.split('/')) == 2:
            day = int(CoolSeasonEnd.split('/')[0])
            month = int(CoolSeasonEnd.split('/')[1])
            from datetime import date
            day_of_year = date(year=2007, month=month, day=day).timetuple().tm_yday
            CoolSeasonEnd = day_of_year



    SetInputData.Program_Line_1 = 'set ComfStand = ' + str(ComfStand)
    SetInputData.Program_Line_2 = 'set CAT = ' + str(CAT)
    SetInputData.Program_Line_3 = 'set ComfMod = ' + str(ComfMod)
    SetInputData.Program_Line_4 = 'set HVACmode = ' + str(HVACmode)
    SetInputData.Program_Line_5 = 'set VentCtrl = ' + str(VentCtrl)
    SetInputData.Program_Line_6 = 'set VSToffset = ' + str(VSToffset)
    SetInputData.Program_Line_7 = 'set MinOToffset = ' + str(MinOToffset)
    SetInputData.Program_Line_8 = 'set MaxWindSpeed = ' + str(MaxWindSpeed)
    SetInputData.Program_Line_9 = 'set ACSTtol = ' + str(-ASTtol)
    SetInputData.Program_Line_10 = 'set AHSTtol = ' + str(ASTtol)
    SetInputData.Program_Line_11 = 'set CoolSeasonStart = ' + str(CoolSeasonStart)
    SetInputData.Program_Line_12 = 'set CoolSeasonEnd = ' + str(CoolSeasonEnd)

    SetComfTemp.Program_Line_2 = f'set ComfTemp = PMOT*{str(CustAST_m)}+{str(CustAST_n)}'

    SetAppLimits.Program_Line_2 = f'set ACSTaul = {str(CustAST_ACSTaul)}'
    SetAppLimits.Program_Line_3 = f'set ACSTall = {str(CustAST_ACSTall)}'
    SetAppLimits.Program_Line_4 = f'set AHSTaul = {str(CustAST_AHSTaul)}'
    SetAppLimits.Program_Line_5 = f'set AHSTall = {str(CustAST_AHSTall)}'

    SetAST.Program_Line_1 = 'set SetpointAcc = ' + str(SetpointAcc)
    SetAST.Program_Line_2 = 'set m = ' + str(CustAST_m)
    SetAST.Program_Line_3 = 'set n = ' + str(CustAST_n)

    SetVOFinputData.Program_Line_1 = 'set MaxTempDiffVOF = ' + str(MaxTempDiffVOF)
    SetVOFinputData.Program_Line_2 = 'set MinTempDiffVOF = ' + str(MinTempDiffVOF)
    SetVOFinputData.Program_Line_3 = 'set MultiplierVOF = ' + str(MultiplierVOF)

    ApplyCAT.Program_Line_1 = 'set CATcoolOffset = ' + str(CATcoolOffset)
    ApplyCAT.Program_Line_2 = 'set CATheatOffset = ' + str(CATheatOffset)
    ApplyCAT.Program_Line_4 = f'set ACSToffset = {str(CustAST_ACSToffset)} + {str(CATcoolOffset)}'
    ApplyCAT.Program_Line_5 = f'set AHSToffset = {str(CustAST_AHSToffset)} + {str(CATheatOffset)}'


def modify_param(idf, parameter, value):

    SetInputData = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                     program.Name == 'SetInputData'][0])
    SetVOFinputData = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
                        program.Name == 'SetVOFinputData'][0])
    SetAST = ([program for program in idf.idfobjects['EnergyManagementSystem:Program'] if
               program.Name == 'SetAST'][0])

    valid_params = [
        'ComfStand',
        'CAT',
        'ComfMod',
        'SetpointAcc',
        'CoolSeasonStart',
        'CoolSeasonEnd',
        'HVACmode',
        'VentCtrl',
        'MaxTempDiffVOF',
        'MinTempDiffVOF',
        'MultiplierVOF',
        'VSToffset',
        'MinOToffset',
        'MaxWindSpeed',
        'ASTtol',
    ]

    if parameter.lower() not in [i.lower() for i in valid_params]:
        raise ValueError(f'The {parameter} parameter you have entered is not valid. Valid parameters are {valid_params}')

    #todo WIP




from accim.sim import accis
import os
try:
    accis.addAccis(
        ScriptType='vrf_mm',
        Output_keep_existing=False,
        Output_type='detailed',
        Output_freqs=['timestep', 'hourly'],
        EnergyPlus_version='22.1',
        TempCtrl='temp',
        ComfStand=[15],
        CAT=[80],
        ComfMod=[3],
        HVACmode=[2],
        VentCtrl=[0],
        VSToffset=[0],
        MinOToffset=[50],
        MaxWindSpeed=[50],
        ASTtol_steps=0.1,
        ASTtol_start=0.1,
        ASTtol_end_input=0.1,
        confirmGen=True
    )
except KeyError:
    print('\nkey error')
    for i in [j for j in os.listdir() if j.endswith('_pymod.idf')]:
        os.remove(i)



# accim - Adaptive-Comfort-Control-Implemented Model
# Copyright (C) 2021-2025 Daniel Sánchez-García

# accim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# accim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from accim.parametric_and_optimisation.funcs_for_besos import param_accis, param_apmv

accim_predef_model_params = {
    # accim predefined models parameters
    'ComfStand': param_accis.modify_ComfStand,
    'CAT': param_accis.modify_CAT,
    'CATcoolOffset': param_accis.modify_CATcoolOffset,
    'CATheatOffset': param_accis.modify_CATheatOffset,
    'ComfMod': param_accis.modify_ComfMod,
    'SetpointAcc': param_accis.modify_SetpointAcc,
    'CoolSeasonStart': param_accis.modify_CoolSeasonStart,
    'CoolSeasonEnd': param_accis.modify_CoolSeasonEnd,
    'HVACmode': param_accis.modify_HVACmode,
    'VentCtrl': param_accis.modify_VentCtrl,
    'MaxTempDiffVOF': param_accis.modify_MaxTempDiffVOF,
    'MinTempDiffVOF': param_accis.modify_MinTempDiffVOF,
    'MultiplierVOF': param_accis.modify_MultiplierVOF,
    'VSToffset': param_accis.modify_VSToffset,
    'MinOToffset': param_accis.modify_MinOToffset,
    'MaxWindSpeed': param_accis.modify_MaxWindSpeed,
    'ASTtol': param_accis.modify_ASTtol,
}
accim_custom_model_params = {
    # accim custom models parameters
    'CustAST_ACSTaul': param_accis.modify_CustAST_ACSTaul,
    'CustAST_ACSTall': param_accis.modify_CustAST_ACSTall,
    'CustAST_AHSTaul': param_accis.modify_CustAST_AHSTaul,
    'CustAST_AHSTall': param_accis.modify_CustAST_AHSTall,
    'CustAST_ASTaul': param_accis.modify_CustAST_ASTaul,
    'CustAST_ASTall': param_accis.modify_CustAST_ASTall,
    'CustAST_m': param_accis.modify_CustAST_m,
    'CustAST_n': param_accis.modify_CustAST_n,
    'CustAST_ACSToffset': param_accis.modify_CustAST_ACSToffset,
    'CustAST_AHSToffset': param_accis.modify_CustAST_AHSToffset,
    'CustAST_ASToffset': param_accis.modify_CustAST_ASToffset,
}
apmv_setpoints_params = {
    # apmv setpoints parameters
    'Adaptive coefficient': param_apmv.change_adaptive_coeff_all_zones,
    'Adaptive cooling coefficient': param_apmv.change_adaptive_coeff_cooling_all_zones,
    'Adaptive heating coefficient': param_apmv.change_adaptive_coeff_heating_all_zones,
    'PMV setpoint': param_apmv.change_pmv_setpoint_all_zones,
    'PMV cooling setpoint': param_apmv.change_pmv_cooling_setpoint_all_zones,
    'PMV heating setpoint': param_apmv.change_pmv_heating_setpoint_all_zones,
}

all_params = {}
all_params.update(accim_predef_model_params)
all_params.update(accim_custom_model_params)
all_params.update(apmv_setpoints_params)


#!/usr/bin/env python
# coding: utf-8

# # Using the class Table() for data analysis

# Say we have previously renamed the EPW files with rename_epw_files(), and the EPWs we currently have are:
# 
# - United-Kingdom_Aberdeen_Present
# 
# - United-Kingdom_London_Present
# 
# We also have transformed the IDF file TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220.idf, and the IDFs we currently have are:
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_1[CM_0[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_1[CM_3[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_2[CM_0[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_2[CM_3[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_3[CM_0[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# - TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220[CS_INT EN16798[CA_3[CM_3[HM_2[VC_0[VO_0[MT_50[MW_50[AT_0.1[NS_X.idf
# 
# Next, we simulate these files with runEp(), so the CSVs we get are:

# In[1]:


import os
print(*[i for i in os.listdir() if i.endswith('.csv') and 'TestModel_onlyGeometryForVRFsystem_2zones_CalcVent_V2220' in i], sep='\n')


# So now, let's use the class Table() to analyse the data.

# ## 1. Getting the full dataset of simulation results

# In[2]:


from accim.data.postprocessing.main import Table
dataset_hourly = Table(
    #datasets=list #Since we are not specifying any list, it will use all available CSVs in the folder
    source_frequency='hourly', # This lets accim know which is the frequency of the input CSVs. Input CSVs with multiple frequencies are also allowed. It can be 'hourly', 'daily', 'monthly' and 'runperiod'. It can also be 'timestep' but might generate errors.
    frequency='hourly', # If 'daily', accim will aggregate the rows in days. It can be 'hourly', 'daily', 'monthly' and 'runperiod'. It can also be 'timestep' but might generate errors.
    frequency_agg_func='sum', #this makes the sum or average when aggregating in days, months or runperiod; since the original CSV frequency is in hour, it won't make any aeffect
    standard_outputs=True, 
    level=['building'], # A list containing the strings 'block' and/or 'building'. For instance, if ['block', 'building'], accim will generate new columns to sum up or average in blocks and building level.
    level_agg_func=['sum', 'mean'], # A list containing the strings 'sum' and/or 'mean'. For instance, if ['sum', 'mean'], accim will generate the new columns explained in the level argument by summing and averaging.
    level_excluded_zones=[],
    #match_cities=bool #Only used when EPW file has NOT been previously renamed
    #manage_epw_names=bool #Only used when EPW file has NOT been previously renamed
    split_epw_names=True, #to split EPW names based on the pattern Country_City_RCPscenario-Year
)

##
# A Table object has been created, which includes a number of accessible variables. The most important is the Pandas Dataframe created from the CSVs, which can be accessed with .df (in this case, dataset_hourly.df)

# In[3]:


dataset_hourly.df


# In[4]:


dataset_hourly.df.shape


# Also, you can see the block list that has been extracted from the zones in the CSV columns, as well as the zones with heating or cooling energy consumption, and the occupied zones:

# In[5]:


dataset_hourly.block_list


# In[6]:


dataset_hourly.hvac_zone_list


# In[7]:


dataset_hourly.occupied_zone_list

##
# In case of lots of CSVs, using the Table class would mean gathering the CSVs to make the Dataframe each time, and it might be very time-consuming. Therefore, the CSVs can be concatenated into a single CSV and store in the same folder with the argument concatenated_csv_name. Afterwards, that concatenated CSV can be read using the argument source_concatenated_csv_filepath each time a Table class object is created.

# In[8]:


concatenated_csvname = 'notebook_example'
dataset_hourly_indirect = Table(
    #datasets=list #Since we are not specifying any list, it will use all available CSVs in the folder
    source_frequency='hourly', # This lets accim know which is the frequency of the input CSVs. Input CSVs with multiple frequencies are also allowed. It can be 'hourly', 'daily', 'monthly' and 'runperiod'. It can also be 'timestep' but might generate errors.
    frequency='hourly', # If 'daily', accim will aggregate the rows in days. It can be 'hourly', 'daily', 'monthly' and 'runperiod'. It can also be 'timestep' but might generate errors.
    frequency_agg_func='sum', #this makes the sum or average when aggregating in days, months or runperiod; since the original CSV frequency is in hour, it won't make any aeffect
    standard_outputs=True, 
    concatenated_csv_name=concatenated_csvname #Useful when working with large datasets. It saves the output dataset to a CSV file, so you don't need to re-do some work. Afterwards, it can be imported with source_concatenated_csv_filepath argument.
    # All other arguments won't have any effect on the output. This is only to store the csv at early stage. Datawrangling will be performed later.
    
)


# We just generated a CSV, whose name begins with 'notebook example', specifies the input data and ends with 'CSVconcatenated'. If there is some row with NaN, it will be generated another CSV reporting this, but ending with 'Rows_with_NaNs'. We should be able to see it below:

# In[9]:


import os
os.listdir()


# Now, let's continue working with the CSVconcatenated we have generated:

# In[10]:


dataset_hourly_indirect = Table(
    source_concatenated_csv_filepath='notebook_example[srcfreq-hourly[freq-hourly[frequency_agg_func-sum[standard_outputs-True[CSVconcatenated.csv',
    # source_frequency='hourly', # The new source frequency is the previous already computed frequency, since we have already aggregated rows based on the previous frequency. Therefore, there is no need to specify this argument.
    # 3 arguments below have been previously specified, and in fact, are stored at the concatenated CSV filename. So you don't need to specify these again.
    # frequency='hourly',
    # frequency_agg_func='sum',
    # standard_outputs=True,
    level=['building'],
    level_agg_func=['sum', 'mean'],
    split_epw_names=True, #to split EPW names based on the format Country_City_RCPscenario-YEar
    
)


# In[11]:


dataset_hourly_indirect.df.shape

##
# You can see we have obtained the same full dataset. Now, let's delete the concatenated CSV file to save some space.

# In[12]:


for i in os.listdir():
    if concatenated_csvname in i:
        os.remove(i)

##
# ## 2. Filtering results

# We have used any of the 2 methods to obtain the full dataset of simulation results.

# At this point, we should have decided what table or graph we intend to make, and therefore we should know what data we are going to show in it. In this case, we want to make a figure with different subplots, showing the different ComfMod values on the rows and the different EPW_City_or_subcountry values on the columns, and in each subplot, a scatter plot with the indoor operative temperature and adaptive heating and cooling setpoints on the y main axis (left spine), and on the secondary y axis, 2 different spines, one of them to show the heating and cooling energy demand, and the other one the air changes. However, let's say we just want to see the subplots for Category CA_3.

# In[13]:


print(*dataset_hourly.df.columns, sep='\n')


# Now, let's extract the data we are really interested in. For this, we can use the method called format_table():

# In[13]:


dataset_hourly.format_table(
    type_of_table='custom', # Used to choose some predefined tables. It can be 'energy demand', 'comfort hours', 'temperature', 'all' or 'custom'
    custom_cols=[ #if type_of_table is 'custom', custom_cols is used to filter the desired columns to keep
        'Adaptive Cooling Setpoint Temperature_No Tolerance (°C)',
        'Adaptive Heating Setpoint Temperature_No Tolerance (°C)',
        'Building_Total_Zone Operative Temperature (°C) (mean)',
        'BLOCK1:ZONE2_EN16798-1 Running mean outdoor temperature (°C)',
        'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
        'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
        'Building_Total_AFN Zone Infiltration Air Change Rate (ach) (summed)'
    ]
)


# So, we have already filtered the columns we wanted. Let's look at the result:

# In[14]:


dataset_hourly.df


# As you can see, the number of columns have been reduced to 29, which are 22 columns for the index and categorical data, and the 7 columns we filtered.

# ## 3. Making graphs

# We will make the figure with scatter_plot(), however it needs some input data we will previously generate with generate_fig_data().

# ### 3.1 Scatter plot:

# As said above, we first need to generate the dataset lists with generate_fig_data():

# In[15]:
print(*dataset_hourly.df.columns, sep='\n')
##
dataset_hourly.generate_fig_data(
    vars_to_gather_rows=['ComfMod', 'Category'], # variables to gather in rows of subplots
    vars_to_gather_cols=['EPW_City_or_subcountry'],# variables to gather in columns of subplots
    detailed_rows=['CM_0[CA_3', 'CM_3[CA_3'], #we only want to see those combinations
    data_on_x_axis='BLOCK1:ZONE2_EN16798-1 Running mean outdoor temperature (°C)', #column name (string) for the data on x axis
    data_on_y_main_axis=[ #list which includes the name of the axis on the first place, and then in the second place, a list which includes the column names you want to plot
        [
            'Indoor Operative Temperature (°C)',
            [
                'Adaptive Cooling Setpoint Temperature_No Tolerance (°C)',
                'Adaptive Heating Setpoint Temperature_No Tolerance (°C)',
                'Building_Total_Zone Operative Temperature (°C) (mean)',
            ]
        ],
    ],
    data_on_y_sec_axis=[ # similarly to above, a list including the name of the secondary y-axis and the column names you want to plot in it
        [
            'Energy (kWh/m2)',
            [
                'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
                'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
            ]
        ],
        [
            'Air renovation (ach)',
            [
                'Building_Total_AFN Zone Infiltration Air Change Rate (ach) (summed)'
            ]
        ]
    ],
    colorlist_y_main_axis=[
        [
            'Indoor Operative Temperature (°C)',
            [
                'b',
                'r',
                'g',
            ]
        ],
    ],
    colorlist_y_sec_axis=[
        [
            'Energy (kWh/m2)',
            [
                'cyan',
                'orange',
            ]
        ],
        [
            'Air renovation (ach)',
            [
                'yellow'
            ]
        ]
    ]
)
##

# And finally we can plot the figure:

# In[16]:


dataset_hourly.scatter_plot(
    supxlabel='Running Mean Outdoor Temperature (°C)', # data label on x axis
    figname=f'WIP_scatterplot_RMOT',
    figsize=6,
    ratio_height_to_width=0.33,
    confirm_graph=True
)


# ### 3.2 Adaptive vs Static data scatter plot:

# A very specific type of scatter plot can be done to show the relationship between data related to adaptive and static setpoint temperatures. In this case, you would need to use the following arguments in generate_fig_data: data_on_y_axis_baseline_plot, baseline and colorlist_baseline_plot_data; finally, you would need to use the scatter_plot_adap_vs_stat() function.

# In[17]:


dataset_hourly.generate_fig_data(
    vars_to_gather_rows=['EPW_City_or_subcountry'], #you can enter multiple variables, for example: ['EPW_City_or_subcountry', 'EPW_Scenario-Year']
    vars_to_gather_cols=['ComfMod', 'Category'], #you can enter multiple variables
    data_on_y_axis_baseline_plot=[ # in this case, you only need to specify a list which includes the data columns you want to plot
        'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
        'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
    ],
    baseline='CM_0[CA_1', # the baseline needs to be in vars_to_gather_cols, and it's going to be shown on x axis. Given the variables we have simulated, you can choose between 'CM_0[CA_1', 'CM_0[CA_2', 'CM_0[CA_3', 'CM_3[CA_1', 'CM_3[CA_2' and 'CM_3[CA_3'
    colorlist_baseline_plot_data=[
        'b',
        'r',
    ]
)

dataset_hourly.scatter_plot_adap_vs_stat(
    supxlabel='Static Energy Demand (kWh/m2)',
    supylabel='Adaptive Energy Demand (kWh/m2)',
    figname='WIP_scatterplot_adap_vs_stat',
    figsize=3,
    confirm_graph=True
)


# Now, let's delete all the images we have created to save some space:

# In[ ]:


for i in os.listdir():
    if i.endswith('.png'):
        os.remove(i)


# ## 4. Making tables

# Now, let's make some tables comparing results. For example, let's make a table to show monthly values of heating and cooling energy demand. First, let's generate the dataset with monthly frequency and let's filter the columns to keep heating and cooling demand at building level.

# In[ ]:


from accim.data.data_postprocessing import Table
dataset_monthly = Table(
    #datasets=list Since we are not specifying any list, it will use all available CSVs in the folder
    source_frequency='hourly',
    frequency='monthly',
    frequency_agg_func='sum', #this makes the sum or average when aggregating in days, months or runperiod; since the original CSV frequency is in hour, it won't make any aeffect
    standard_outputs=True,
    level=['building'],
    level_agg_func=['sum'],
    #match_cities=bool Only used when EPW file has NOT been previously renamed with the function rename_epw_files
    #manage_epw_names=bool Only used when EPW file has NOT been previously renamed with the function rename_epw_files
    split_epw_names=True, #to split EPW names based on the format Country_City_RCPscenario-YEar
)

dataset_monthly.format_table(
    type_of_table='custom',
    custom_cols=[
        'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
        'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
    ]
)


# Now, let's make the table we are looking for. You can choose 'unstack' or 'pivot' as a reshaping method (remember that we are actually working with Pandas Dataframes). Let's go with unstack, and we are going to compare the baseline shown below with all gathered variables

# In[ ]:


dataset_monthly.wrangled_table(
    reshaping='unstack', #can be 'unstack' or 'pivot'
    vars_to_gather=['ComfMod', 'Category'],
    baseline='CM_0[CA_1',
    comparison_mode='baseline compared to others', #can be 'baseline compared to others' or 'others compared to baseline'
    comparison_cols=['relative', 'absolute'] #'relative' to show the difference as a percentage, 'absolute' to show the difference by subtracting
)


# In[ ]:


dataset_monthly.wrangled_df_unstacked


# Let's make a different dataset to work with runperiod frequency.

# In[ ]:


from accim.data.data_postprocessing import Table
dataset_runperiod = Table(
    #datasets=list Since we are not specifying any list, it will use all available CSVs in the folder
    source_frequency='hourly',
    frequency='runperiod',
    frequency_agg_func='sum', #this makes the sum or average when aggregating in days, months or runperiod; since the original CSV frequency is in hour, it won't make any aeffect
    standard_outputs=True,
    level=['building'],
    level_agg_func=['sum'],
    #match_cities=bool Only used when EPW file has NOT been previously renamed
    #manage_epw_names=bool Only used when EPW file has NOT been previously renamed
    split_epw_names=True, #to split EPW names based on the format Country_City_RCPscenario-YEar
)

dataset_runperiod.format_table(
    type_of_table='custom',
    custom_cols=[
        'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
        'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
    ]
)


# In this case, we are going to use the 'pivot' reshaping option:

# In[ ]:


dataset_runperiod.wrangled_table(
    reshaping='pivot',
    vars_to_gather=['ComfMod', 'Category'],
    baseline='CM_0[CA_1',
    comparison_mode='baseline compared to others',
    comparison_cols=['relative', 'absolute']
)


# In[ ]:


dataset_runperiod.wrangled_df_pivoted


# Again, please remember we are working with Pandas Dataframe objects. That means we can modify the df we have generated. For instance, we can simplify it by removing all rows except those with 'CA_3' in the Category column, as shown below.

# In[ ]:


dataset_runperiod.df = dataset_runperiod.df[
    dataset_runperiod.df['Category'].isin(['CA_3'])
]


# In[ ]:


dataset_runperiod.df


# And now we could make a similar pivoted table as above, but simplified to show only CA_3:

# In[ ]:


dataset_runperiod.wrangled_table(
    reshaping='pivot',
    vars_to_gather=['ComfMod'],
    baseline='CM_0',
    comparison_mode='baseline compared to others',
    comparison_cols=['relative', 'absolute']
)


# In[ ]:


dataset_runperiod.wrangled_df_pivoted


# Another method to achieve this is generating a different Table object, only reading the CSV files with CA_3 in it name:

# In[ ]:


import os
from accim.data.data_postprocessing import Table

dataset = [i for i in os.listdir() if i.endswith('.csv') and 'CA_3' in i]

dataset_runperiod_simplified_1 = Table(
    datasets=dataset,
    source_frequency='hourly',
    frequency='runperiod',
    frequency_agg_func='sum',
    # this makes the sum or average when aggregating in days, months or runperiod; since the original CSV frequency is in hour, it won't make any aeffect
    standard_outputs=True,
    level=['building'],
    level_agg_func=['sum'],
    # match_cities=bool Only used when EPW file has NOT been previously renamed
    # manage_epw_names=bool Only used when EPW file has NOT been previously renamed
    split_epw_names=True,  # to split EPW names based on the format Country_City_RCPscenario-YEar
)

dataset_runperiod_simplified_1.format_table(
    type_of_table='custom',
    custom_cols=[
        'Building_Total_Cooling Energy Demand (kWh/m2) (summed)',
        'Building_Total_Heating Energy Demand (kWh/m2) (summed)',
    ]
)

dataset_runperiod_simplified_1.wrangled_table(
    reshaping='pivot',
    vars_to_gather=['ComfMod'],
    baseline='CM_0',
    comparison_mode='baseline compared to others',
    comparison_cols=['relative', 'absolute']
)




# You can see below the resulting table is exactly the same:

# In[ ]:


dataset_runperiod_simplified_1.wrangled_df_pivoted


# You can finally export the table as xlsx format:

# In[ ]:


dataset_runperiod_simplified_1.wrangled_df_pivoted.to_excel('building_energy_demand.xlsx')


# Let's delete it now:

# In[ ]:


for i in os.listdir():
    if i.endswith('.xlsx'):
        os.remove(i)


# In[ ]:







# accim - Adaptive-Comfort-Control-Implemented Model
# Copyright (C) 2021-2025 Daniel Sánchez-García

# accim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# accim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

def average_results(result):
    return result.data["Value"].mean()
def sum_results(result):
    return result.data["Value"].sum()
def return_time_series(result):
    return result.data["Value"].to_list()



from accim.sim.accis import addAccis

x = addAccis(
    ScriptType='vrf_mm',
    SupplyAirTempInputMethod='supply air temperature',
    Output_keep_existing=False,
    Output_type='standard',
    Output_freqs=['hourly'],
    Output_gen_dataframe=False,
    EnergyPlus_version='23.2',
    TempCtrl='temp',

)


# accim - Adaptive-Comfort-Control-Implemented Model
# Copyright (C) 2021-2025 Daniel Sánchez-García

# accim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# accim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Submodule to perform data processing before simulation run.
"""
import geopy
from geopy import Nominatim

def capitalize_words(text):
  """
  Capitalizes the first letter of each word in a string.

  Args:
    text: The input string.

  Returns:
    A new string with the first letter of each word capitalized.
  """
  words = text.split()
  capitalized_words = [word.capitalize() for word in words]
  new_string = " ".join(capitalized_words)
  new_string = new_string.replace(" ", "-")
  return new_string

class give_address_ssl:
    """
    Search address from coordinates using nominatim.openstreetmap.org.
    Version 1 to avoid ssl error.

    :param latitude: The latitude from the EPW file.
    :type latitude: float
    :param longitude: The longitude from the EPW file.
    :type longitude: float
    """
    def __init__(
            self,
            latitude: float,
            longitude: float,
    ):
        """
        Constructor method.
        """
        import requests
        import ssl
        import certifi

        # Set up SSL context with OpenStreetMap certificate added to trust store
        context = ssl.create_default_context()
        cafile = certifi.where()
        context.load_verify_locations(cafile)
        context.check_hostname = True

        # Make request to OpenStreetMap API with SSL verification enabled
        url = f"https://nominatim.openstreetmap.org/reverse?format=jsonv2&lat={latitude}&lon={longitude}"
        response = requests.get(url, verify=cafile, headers={'User-Agent': 'Mozilla/5.0'}, timeout=10).json()

        # Extract address information from response
        self.address = response['address']

        # Print the full address
        self.full_address = ", ".join([v for k,v in self.address.items() if v and k != 'country_code'])

class give_address_openssl:
    """
    Search address from coordinates using nominatim.openstreetmap.org.
    Version 2 to avoid ssl error.

    :param latitude: The latitude from the EPW file.
    :type latitude: float
    :param longitude: The longitude from the EPW file.
    :type longitude: float
    """
    def __init__(
            self,
            latitude: float,
            longitude: float,
    ):
        import requests
        import certifi
        import OpenSSL
        import ssl
        import socket

        # Obtain the certificate chain
        ctx = ssl.create_default_context()
        sock = ctx.wrap_socket(socket.socket(), server_hostname="nominatim.openstreetmap.org")
        sock.connect(("nominatim.openstreetmap.org", 443))
        cert = sock.getpeercert(True)
        cert_chain = [OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1, cert)]

        # Set up SSL context with certificate chain added to trust store
        cafile = certifi.where()
        context = ssl.create_default_context(cafile=cafile)
        for cert in cert_chain:
            context.load_verify_locations(cadata=cert.dump())

        # Make request to OpenStreetMap API with SSL verification enabled
        url = f"https://nominatim.openstreetmap.org/reverse?format=jsonv2&lat={latitude}&lon={longitude}"
        response = requests.get(url, verify=cafile, headers={'User-Agent': 'Mozilla/5.0'}, timeout=10).json()

        # Extract address information from response
        self.address = response['address']

        # Print the full address
        self.full_address = ", ".join([v for k,v in self.address.items() if v and k != 'country_code'])


class give_address:
    """
    Search address from coordinates using nominatim.openstreetmap.org.
    Version 3 to avoid ssl error.

    :param latitude: The latitude from the EPW file.
    :type latitude: float
    :param longitude: The longitude from the EPW file.
    :type longitude: float
    """
    def __init__(
            self,
            latitude: float,
            longitude: float,
    ):
        import requests

        # Make request to OpenStreetMap API
        url = f"https://nominatim.openstreetmap.org/reverse?format=jsonv2&lat={latitude}&lon={longitude}"
        response = requests.get(url).json()

        # Extract address information from response
        self.address = response['address']

        # Print the full address
        self.full_address = ", ".join([v for k,v in self.address.items() if v and k != 'country_code'])


class rename_epw_files:
    """
    Renames the EPW files following the name convention 'Country_City_RCPscenario-Year'.
    The Country and City fields are computed based on the coordinates of the EPW,
    and the RCPscenario and Year are taken from the original name.
    If no reference is found, the sample_EPWs are considered to be  for Present scenario.

    :param filelist: A list of the EPW files.
        If omitted, it will rename all sample_EPWs in that folder.
    :type filelist: list
    :param rename_city_dict: A dict to set the city field for each EPW file.
        It must follow the pattern
        {'city name to be search in epw file name': 'city name for the epw file if found`}
    :type rename_city_dict: dict
    :param country_to_city_dict: A dict to set the country field for each EPW file.
        It must follow the pattern
        {'country name': [output city names defined in rename_city_dict]}
    :type country_to_city_dict: dict
    :param confirm_renaming: True or False,
        to skip renaming confirmation on prompt command or console
    :type confirm_renaming: bool
    :param confirm_deletion: True or False,
        to skip deletion confirmation on prompt command or console
    :type confirm_deletion: bool
    """
    def __init__(
            self,
            filelist=None,
            rename_city_dict: dict = None,
            country_to_city_dict: dict = None,
            confirm_renaming=None,
            confirm_deletion=None,
    ):
        """
        Constructor method.
        """
        import glob
        import pandas as pd
        import datapackage
        import numpy as np
        import os
        # from geopy.exc import GeocoderUnavailable
        # from geopy.geocoders import Nominatim
        import pycountry
        import shutil
        from unidecode import unidecode
        from time import time
        import requests

        checkpoint = 0

        start = time()

        if filelist is None:
            filelist = []

        if len(filelist) > 0:
            epw_files_to_rename = filelist
        else:
            epw_files_to_rename = glob.glob('*.epw', recursive=True)

        epw_df = pd.DataFrame(data=epw_files_to_rename,
                              index=list(range(len(epw_files_to_rename))),
                              columns=['EPW_file_names'])

        checkpoint = checkpoint + 1

        for i in range(len(epw_df)):
            epw_df.loc[i, 'EPW_abs_path'] = os.path.abspath(epw_df.loc[i, 'EPW_file_names'])

        epw_df['EPW_names'] = epw_df['EPW_file_names'].str.replace('.epw', '')

        epw_df['EPW_mod'] = epw_df['EPW_names'].str.replace('-', '_').str.replace('.', '_').str.split('_')

        epw_df['EPW_mod_filtered'] = epw_df['EPW_names'].str.replace('-', '_').str.replace('.', '_').str.split('_')

        future_scenario_dict = {
            'Present': ['Presente', 'Actual', 'Present', 'Current'],
            'RCP26': ['RCP2.6', 'RCP26', 'RCP2.5', 'RCP25'],
            'RCP45': ['RCP4.5', 'RCP45'],
            'RCP60': ['RCP6.0', 'RCP60'],
            'RCP85': ['RCP8.5', 'RCP85'],
            'SSP126': ['ssp126'],
            'SSP245': ['ssp245'],
            'SSP370': ['ssp370'],
            'SSP585': ['ssp585'],
        }

        for i in range(len(epw_df['EPW_names'])):
            for j in future_scenario_dict:
                for k in future_scenario_dict[j]:
                    if k.lower() in epw_df.loc[i, 'EPW_names'].lower():
                        epw_df.loc[i, 'EPW_scenario'] = j
                        try:
                            epw_df.loc[i, 'EPW_mod_filtered'].remove([x for x in epw_df.loc[i, 'EPW_mod_filtered'] if x.lower() in k.lower()][0])
                        except IndexError:
                            continue

        future_scenario_not_found_list = []

        for i in range(len(epw_df['EPW_names'])):
            if epw_df.loc[i, 'EPW_scenario'] == 'nan':
                epw_df.loc[i, 'EPW_scenario'] = 'Present'
                future_scenario_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])
            # try:
            #     if type(epw_df.loc[i, 'EPW_scenario']) is float:
            #         epw_df.loc[i, 'EPW_scenario'] = 'Present'
            #         future_scenario_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])
            # except KeyError:
            #     epw_df.loc[i, 'EPW_scenario'] = 'Present'
            #     future_scenario_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])

        if len(future_scenario_not_found_list) > 0:
            print('Since no match has been found between RCP or SSP scenarios and EPW file name, '
                  'Present scenario has been assigned to the following EPW files:')
            print(*future_scenario_not_found_list, sep='\n')

        for i in range(len(epw_df['EPW_names'])):
            for j in range(2000, 2101, 10):
                if str(j) in epw_df.loc[i, 'EPW_names']:
                    try:
                        epw_df.loc[i, 'EPW_year'] = str(j)
                        epw_df.loc[i, 'EPW_mod_filtered'].remove(str(j))
                    except ValueError:
                        continue
                        # print('Year has not been identified.')

        for i in range(len(epw_df['EPW_names'])):
            if epw_df.loc[i, 'EPW_scenario'] == 'Present':
                epw_df.loc[i, 'EPW_scenario_year'] = 'Present'
            else:
                epw_df.loc[i, 'EPW_scenario_year'] = epw_df.loc[i, 'EPW_scenario'] + '-' + epw_df.loc[i, 'EPW_year']

        year_not_found_list = []
        for i in range(len(epw_df['EPW_names'])):
            if epw_df.loc[i, 'EPW_year'] == 'nan':
                epw_df.loc[i, 'EPW_year'] = 'Present'
                year_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])
            # try:
            #     if type(epw_df.loc[i, 'EPW_year']) is float:
            #         epw_df.loc[i, 'EPW_year'] = 'Present'
            #         year_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])
            # except KeyError:
            #     epw_df.loc[i, 'EPW_year'] = 'Present'
            #     year_not_found_list.append(epw_df.loc[i, 'EPW_file_names'])
        if len(year_not_found_list) > 0:
            print('Since no match has been found between RCP or SSP scenario Year and EPW file name, '
                  'Present year has been assigned to the following EPW files:')
            print(*year_not_found_list, sep='\n')

        checkpoint = checkpoint + 1

        # path = r'C:\Users\user\PycharmProjects\accim'
        path = os.getcwd()
        new_list = []
        for file in epw_files_to_rename:
            # open the file and then call .read() to get the text
            with open(
                    os.path.join(path, file),
                    "rb") as f:
                text = f.readline()
                new_list.append([str(text).split(',')])
            # f.close()

        for i in range(len(new_list)):
            epw_df.loc[i, 'EPW_latitude'] = new_list[i][0][6]
            epw_df.loc[i, 'EPW_longitude'] = new_list[i][0][7]

        checkpoint = checkpoint + 1

        epw_df['EPW_city_or_subcountry'] = 'UNKNOWN'
        epw_df['EPW_country'] = 'UNKNOWN'

        for i in range(len(epw_df)):
            if rename_city_dict is not None:
                for city in rename_city_dict:
                    if city.lower() in epw_df.loc[i, 'EPW_names'].lower():
                        epw_df.loc[i, 'EPW_city_or_subcountry'] = capitalize_words(rename_city_dict[city])
            if country_to_city_dict is not None:
                for country in country_to_city_dict:
                    for cities in country_to_city_dict[country]:
                        if capitalize_words(cities).lower() == epw_df.loc[i, 'EPW_city_or_subcountry'].lower():
                            epw_df.loc[i, 'EPW_country'] = capitalize_words(country)
            if epw_df.loc[i, 'EPW_city_or_subcountry'] == 'UNKNOWN' or epw_df.loc[i, 'EPW_country'] == 'UNKNOWN':
                try:
                    geolocator = Nominatim(user_agent="abcd")
                    location = geolocator.reverse(epw_df.loc[i, 'EPW_latitude'] + "," + epw_df.loc[i, 'EPW_longitude'])
                    # address = location.raw['address']

                    epw_df.loc[i, 'EPW_country_code'] = location.raw['address']['country_code'].upper()
                    epw_df.loc[i, 'EPW_country'] = pycountry.countries.get(alpha_2=epw_df.loc[i, 'EPW_country_code']).name.replace(' ', '-')
                    # epw_df.loc[i, 'EPW_country'] = capitalize_words(location.raw['address']['country'])
                    epw_df.loc[i, 'location_address'] = unidecode(location.raw['display_name'])

                    loc_matches = []
                    for j in epw_df.loc[i, 'EPW_mod_filtered']:
                        if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                            if j != 'IWEC' and j != 'SWEC':
                                if j in location.raw['display_name']:
                                    loc_matches.append(j)
                    proposed_city = capitalize_words(' '.join(loc_matches))
                    epw_df.loc[i, 'EPW_city_or_subcountry'] = proposed_city

                                # if any(j.lower() in [c for c in str(unidecode(location.raw['address']['city'])).lower().split(' ')]):
                                #     epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                            # try:
                            #     if j.lower() in str(unidecode(osm_address.address['city_district'])).lower():
                            #         epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                            # except KeyError:
                            #     pass

                except geopy.exc.GeocoderUnavailable:
                    try:
                        osm_address = give_address(epw_df.loc[i, 'EPW_latitude'], epw_df.loc[i, 'EPW_longitude'])

                        epw_df.loc[i, 'EPW_country_code'] = osm_address.address['country_code'].upper()
                        epw_df.loc[i, 'EPW_country'] = pycountry.countries.get(alpha_2=epw_df.loc[i, 'EPW_country_code']).name.replace(' ', '-')
                        epw_df.loc[i, 'location_address'] = unidecode(osm_address.full_address)

                        for j in epw_df.loc[i, 'EPW_mod_filtered']:
                            if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                try:
                                    if j.lower() in str(unidecode(osm_address.address['city_district'])).lower():
                                        epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                except KeyError:
                                    pass
                            if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                try:
                                    if j.lower() in str(unidecode(osm_address.address['city'])).lower():
                                        epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                except KeyError:
                                    pass
                            if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                try:
                                    if j.lower() in str(unidecode(osm_address.address['municipality'])).lower():
                                        epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                except KeyError:
                                    pass
                    except requests.exceptions.SSLError:
                        try:
                            osm_address = give_address_ssl(epw_df.loc[i, 'EPW_latitude'], epw_df.loc[i, 'EPW_longitude'])

                            epw_df.loc[i, 'EPW_country_code'] = osm_address.address['country_code'].upper()
                            epw_df.loc[i, 'EPW_country'] = pycountry.countries.get(alpha_2=epw_df.loc[i, 'EPW_country_code']).name.replace(' ', '-')
                            epw_df.loc[i, 'location_address'] = unidecode(osm_address.full_address)

                            for j in epw_df.loc[i, 'EPW_mod_filtered']:
                                if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                    try:
                                        if j.lower() in str(unidecode(osm_address.address['city_district'])).lower():
                                            epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                    except KeyError:
                                        pass
                                if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                    try:
                                        if j.lower() in str(unidecode(osm_address.address['city'])).lower():
                                            epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                    except KeyError:
                                        pass
                                if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                    try:
                                        if j.lower() in str(unidecode(osm_address.address['municipality'])).lower():
                                            epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                    except KeyError:
                                        pass
                        except requests.exceptions.SSLError:
                            try:
                                osm_address = give_address_openssl(epw_df.loc[i, 'EPW_latitude'], epw_df.loc[i, 'EPW_longitude'])

                                epw_df.loc[i, 'EPW_country_code'] = osm_address.address['country_code'].upper()
                                epw_df.loc[i, 'EPW_country'] = pycountry.countries.get(alpha_2=epw_df.loc[i, 'EPW_country_code']).name.replace(' ', '-')
                                epw_df.loc[i, 'location_address'] = unidecode(osm_address.full_address)

                                for j in epw_df.loc[i, 'EPW_mod_filtered']:
                                    if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                        try:
                                            if j.lower() in str(unidecode(osm_address.address['city_district'])).lower():
                                                epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                        except KeyError:
                                            pass
                                    if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                        try:
                                            if j.lower() in str(unidecode(osm_address.address['city'])).lower():
                                                epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                        except KeyError:
                                            pass
                                    if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                                        try:
                                            if j.lower() in str(unidecode(osm_address.address['municipality'])).lower():
                                                epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
                                        except KeyError:
                                            pass
                            except AttributeError:
                                epw_df.loc[i, 'EPW_country_code'] = 'UNKNOWN'
                                epw_df.loc[i, 'EPW_country'] = 'UNKNOWN'
                                epw_df.loc[i, 'location_address'] = 'UNKNOWN'
                                epw_df.loc[i, 'EPW_city_or_subcountry'] = 'UNKNOWN'
                                print(f"For some reason, accim cannot connect to OpenStreetMap to get the address of file {epw_df.loc[i, 'EPW_names']}. It gets an SSL error.")

        checkpoint +=1

        # geolocator = Nominatim(user_agent="geoapiExercises")
        # for i in range(len(epw_df)):
        #     try:
        #         location = geolocator.reverse(epw_df.loc[i, 'EPW_latitude'] + "," + epw_df.loc[i, 'EPW_longitude'])
        #         # print(location)
        #         epw_df.loc[i, 'EPW_country_code'] = location.raw['address'].get('country_code').upper()
        #         epw_df.loc[i, 'EPW_country'] = pycountry.countries.get(alpha_2=epw_df.loc[i, 'EPW_country_code']).name.replace(' ', '-')
        #         # epw_df.loc[i, 'EPW_country'] = location.raw['address'].get('country').replace(' ', '-')
        #         # epw_df.loc[i, 'EPW_city'] = location.raw['address'].get('city', '')
        #         # epw_df.loc[i, 'geolocator_address'] = location.raw['display_name']
        #         epw_df.loc[i, 'location_address'] = unidecode(location.address)
        #
        #
        #     # for i in range(len(epw_df['EPW_mod_filtered'])):
        #     #     location = geolocator.reverse(epw_df.loc[i, 'EPW_latitude'] + "," + epw_df.loc[i, 'EPW_longitude'])
        #         for j in epw_df.loc[i, 'EPW_mod_filtered']:
        #             if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
        #                 try:
        #                     if j.lower() in str(unidecode(location.raw['address']['city_district'])).lower():
        #                         epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
        #                 except KeyError:
        #                     pass
        #             if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
        #                 try:
        #                     if j.lower() in str(unidecode(location.raw['address']['city'])).lower():
        #                         epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
        #                 except KeyError:
        #                     pass
        #             if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
        #                 try:
        #                     if j.lower() in str(unidecode(location.raw['address']['municipality'])).lower():
        #                         epw_df.loc[i, 'EPW_city_or_subcountry'] = j.replace(' ', '-').capitalize()
        #                 except KeyError:
        #                     pass
        #     except GeocoderUnavailable:
        #         pass

        for i in range(len(epw_df)):
            if epw_df.loc[i, 'EPW_city_or_subcountry'] == 'UNKNOWN' or epw_df.loc[i, 'EPW_country'] == 'UNKNOWN':
                print(f"No exact match for the epw data has been found for epw file {epw_df.loc[i, 'EPW_names']}. The address obtained from the coordinates is:\n{epw_df.loc[i, 'location_address']}")
                if epw_df.loc[i, 'EPW_city_or_subcountry'] == 'UNKNOWN':
                    city_or_subcountry = input('Please enter the city or subcountry: ')
                    epw_df.loc[i, 'EPW_city_or_subcountry'] = capitalize_words(city_or_subcountry)
                if epw_df.loc[i, 'EPW_country'] == 'UNKNOWN':
                    country = input('Please enter the country: ')
                    epw_df.loc[i, 'EPW_country'] = capitalize_words(country)

        checkpoint = checkpoint + 1

        # for i in range(len(epw_df['EPW_mod_filtered'])):
        #     try:
        #         if type(epw_df.loc[i, 'EPW_city_or_subcountry']) is float:
        #             osm_address = give_address(epw_df.loc[i, 'EPW_latitude'], epw_df.loc[i, 'EPW_longitude'])
        #             # location = geolocator.reverse(epw_df.loc[i, 'EPW_latitude'] + "," + epw_df.loc[i, 'EPW_longitude'])
        #             try:
        #                 epw_df.loc[i, 'EPW_city_or_subcountry'] = str(osm_address.address['city'].replace(' ', '-')).capitalize()
        #                 # epw_df.loc[i, 'EPW_city_or_subcountry'] = str(location.raw['address'].get('city').replace(' ', '-')).capitalize()
        #             except AttributeError:
        #                 epw_df.loc[i, 'EPW_city_or_subcountry'] = str(osm_address.address['city']).capitalize()
        #                 # epw_df.loc[i, 'EPW_city_or_subcountry'] = str(location.raw['address'].get('city')).capitalize()
        #         if type(epw_df.loc[i, 'EPW_country']) is float:
        #             osm_address = give_address(epw_df.loc[i, 'EPW_latitude'], epw_df.loc[i, 'EPW_longitude'])
        #             # location = geolocator.reverse(epw_df.loc[i, 'EPW_latitude'] + "," + epw_df.loc[i, 'EPW_longitude'])
        #             try:
        #                 epw_df.loc[i, 'EPW_country'] = osm_address.address['country'].replace(' ', '-').capitalize()
        #                 # epw_df.loc[i, 'EPW_country'] = location.raw['address'].get('country').replace(' ', '-').capitalize()
        #             except AttributeError:
        #                 epw_df.loc[i, 'EPW_country'] = osm_address.address['country'].capitalize()
        #                 # epw_df.loc[i, 'EPW_country'] = location.raw['address'].get('country').capitalize()
        #     except KeyError:
        #         continue

        for col in ['EPW_country', 'EPW_city_or_subcountry', 'EPW_scenario_year']:
            for row in range(len(epw_df)):
                try:
                    if epw_df.loc[row, col] is None:
                        epw_df.loc[row, col] = 'UNKNOWN'
                except KeyError:
                    continue

        checkpoint = checkpoint + 1

        epw_df['EPW_new_names'] = epw_df[['EPW_country', 'EPW_city_or_subcountry', 'EPW_scenario_year']].agg('_'.join, axis=1)

        checkpoint = checkpoint + 1

        end = time()
        time_taken = round(end - start, 2)
        time_per_epw = round(time_taken / len(epw_df), 2)
        print(f'The geolocation process has taken: {time_taken} seconds ({time_per_epw} s/EPW)')

        print('\nThe previous and new names of the EPW files and their unique IDs are:')
        for i in range(len(epw_df)):
            print(f'ID: {i} / {epw_df.loc[i, "EPW_names"]} / {epw_df.loc[i, "EPW_new_names"]}')
        # print(*list(epw_df['EPW_names']), sep='\n')
        # print('And the new names of the EPW files are going to be:')
        # print(*list(epw_df['EPW_new_names']), sep='\n')

        # Checking there are no duplicates
        unique_list = []
        duplicated_list = []
        duplicated_id_list = []
        unknown_id_list = []
        for i in range(len(epw_df)):
            if epw_df.loc[i, 'EPW_new_names'] not in unique_list:
                unique_list.append(epw_df.loc[i, 'EPW_new_names'])
            else:
                duplicated_list.append(epw_df.loc[i, 'EPW_new_names'])
                unique_list.remove(epw_df.loc[i, 'EPW_new_names'])
        for i in range(len(epw_df)):
            if epw_df.loc[i, 'EPW_new_names'] in duplicated_list:
                duplicated_id_list.append(i)
        if len(duplicated_list) > 0:
            duplicated_list = list(set(duplicated_list))
            duplicated_id_list = list(set(duplicated_id_list))
            print('\nDuplicates have been found in the renamed EPW files, '
                  'therefore these need to be amended in the next stage:')
            print(duplicated_list)
        for i in range(len(epw_df)):
            if 'UNKNOWN' in epw_df.loc[i, 'EPW_city_or_subcountry']:
                unknown_id_list.append(i)
        if len(unknown_id_list) > 0:
            print('\n"UNKNOWN" city or subcountry have been found in the renamed EPW files, '
                  'therefore these need to be amended in the next stage.')

        checkpoint = checkpoint + 1

        amendments_list = list(
            int(i)
            for i
            in input('\nIf any of the city or subcountry names needs some amendment '
                     '(if you are not happy with any of the available options, '
                     'you can exclude it from renaming at the next stage), '
                     'please enter the EPW IDs separated by space; otherwise, hit enter to omit:').split()
        )
        while len(list(set([i for i in amendments_list if amendments_list.count(i) > 1]))) > 0:
            amendments_list = list(
                int(i)
                for i
                in input('Some of the EPW IDs you just entered are duplicated. '
                         'Please enter again the epw IDs separated by space:').split()
            )

        for i in duplicated_id_list:
            if i not in amendments_list:
                amendments_list.append(i)
        for i in unknown_id_list:
            if i not in amendments_list:
                amendments_list.append(i)

        amendments_list = sorted(amendments_list)

        if len(amendments_list) > 0:
            epw_df['amended_city_or_subcountry'] = 'temp'
            for i in amendments_list:
                print(f'\nRegarding the file ID: {i} / old name: {epw_df.loc[i, "EPW_names"]} / new name: {epw_df.loc[i, "EPW_new_names"]}, the address obtained from coordinates is: ')
                print(epw_df.loc[i, "location_address"])
                epw_df.loc[i, 'amended_city_or_subcountry'] = input('Please enter the amended city or subcountry, which must be unique: ').replace(' ', '-').replace('_', '-')
                temp_name = f'{epw_df.loc[i, "EPW_country"]}_{epw_df.loc[i, "amended_city_or_subcountry"]}_{epw_df.loc[i, "EPW_scenario_year"]}'
                epw_df.loc[i, 'EPW_new_names'] = temp_name
                while list(epw_df['EPW_new_names']).count(temp_name) > 1:
                    print(f"{epw_df.loc[i, 'EPW_new_names']} already exists in the EPW file list, therefore you need to select a different city or subcountry name.")
                    epw_df.loc[i, 'amended_city_or_subcountry'] = input('Please enter again the amended city or subcountry, which must be unique: ').replace(' ', '-').replace('_', '-')
                    temp_name = f'{epw_df.loc[i, "EPW_country"]}_{epw_df.loc[i, "amended_city_or_subcountry"]}_{epw_df.loc[i, "EPW_scenario_year"]}'
                    epw_df.loc[i, 'EPW_new_names'] = temp_name

            print('\nThe previous and new names of the EPW files after city or subcountry name amendments and their unique IDs are:')
            for i in amendments_list:
                print(f'ID: {i} / {epw_df.loc[i, "EPW_names"]} / {epw_df.loc[i, "EPW_new_names"]}')

        print('\nThe final list of previous and new names of the EPW files and their unique IDs is:')
        for i in range(len(epw_df)):
            print(f'ID: {i} / {epw_df.loc[i, "EPW_names"]} / {epw_df.loc[i, "EPW_new_names"]}')
        exclusion_list = list(
            int(id)
            for id
            in input('\nIf you want to exclude some EPWs from renaming, '
                     'please enter the IDs separated by space, '
                     'otherwise, hit enter to continue:').split()
        )

        if confirm_renaming is None:
            proceed = input('\nDo you want to copy and rename the file or files? [y/n]:')
            if 'y' in proceed:
                confirm_renaming = True
            elif 'n' in proceed:
                confirm_renaming = False

        checkpoint = checkpoint + 1

        if confirm_renaming:
            for i in range(len(epw_df)):
                if i not in exclusion_list:
                    try:
                        shutil.copy(epw_df.loc[i, 'EPW_abs_path'], path + '/' + epw_df.loc[i, 'EPW_new_names'] + '.epw')
                        print(f'The file {epw_df.loc[i, "EPW_names"]} has been renamed to {epw_df.loc[i, "EPW_new_names"]}')
                    except shutil.SameFileError:
                        print(f'The old and new names of file {epw_df.loc[i, "EPW_names"]} are the same.')
                        epw_files_to_rename.remove(epw_df.loc[i, "EPW_names"] + '.epw')

        if confirm_deletion is None:
            proceed = input('\nDo you want to delete the original EPW file or files? [y/n]:')
            if 'y' in proceed:
                confirm_deletion = True
            elif 'n' in proceed:
                confirm_deletion = False

        if confirm_deletion:
            for i in epw_files_to_rename:
                os.remove(i)
                print(f'The file {i} has been deleted.')

        # todo pop up when process ends; by default True
